/*
 * catch.c - implementation of the ICI catch type. This is an internal
 * type used as a marker on the execution stack. These objects should
 * never escape to user visibility. See also catch.h.
 */
#include "exec.h"
#include "catch.h"
#include "op.h"
#include "func.h"

/*
 * Unwind the execution stack until a catcher is found.  Then unwind
 * the scope and operand stacks to the matching depth (but only if it is).
 * Returns the catcher, or NULL if there wasn't one.
 */
catch_t *
ici_unwind()
{
    object_t	**p;
    catch_t	*c;

    for (p = x_top - 1; p >= xs->a_base; --p)
    {
	if (iscatch(*p))
	{
	    c = catchof(*p);
	    x_top = p;
	    o_top = &os->a_base[c->c_odepth];
	    v_top = &vs->a_base[c->c_vdepth];
	    NEXT_VSVER;
	    return c;
	}
    }
    return NULL;
}

/*
 * Mark this and referenced unmarked objects, return memory costs.
 * See comments on t_mark() in object.h.
 */
static unsigned long
mark_catch(object_t *o)
{
    unsigned long	mem;

    o->o_flags |= O_MARK;
    mem = sizeof(catch_t);
    if (catchof(o)->c_catcher != NULL)
	mem += mark(catchof(o)->c_catcher);
    return mem;
}

/*
 * Return a new catch object with the given catcher object and
 * corresponding to the operand and variable stack depths given.
 * The catcher, o, may be NULL.
 *
 * Note: catch's are special. Unlike most types this new_catch function
 * returns its object with a ref count of 0 not 1.
 */
catch_t *
new_catch(object_t *o, int odepth, int vdepth)
{
    register catch_t	*c;

    if ((c = talloc(catch_t)) == NULL)
	return NULL;
    objof(c)->o_type = &ici_catch_type;
    objof(c)->o_tcode = TC_CATCH;
    objof(c)->o_flags = 0;
    objof(c)->o_nrefs = 0;/* Catch's are pre-decref'ed.  Unlike other types. */
    rego(c);
    c->c_catcher = o;
    c->c_odepth = odepth;
    c->c_vdepth = vdepth;
    return c;
}

/*
 * runner handler	=> catcher (os)
 *			=> catcher pc (xs)
 *			=> catcher (vs)
 */
int
ici_op_onerror()
{
    if ((x_top[-1] = objof(new_catch(o_top[-1], o_top - os->a_base - 2, v_top - vs->a_base))) == NULL)
	return 1;
    if ((*x_top = objof(new_pc(arrayof(o_top[-2])))) == NULL)
	return 1;
    o_top -= 2;
    ++x_top;
    return 0;
}

type_t	ici_catch_type =
{
    mark_catch,
    free_simple,
    hash_unique,
    cmp_unique,
    copy_simple,
    assign_simple,
    fetch_simple,
    "catch"
};

op_t	o_onerror	= {OBJ(TC_OP, op_type), ici_op_onerror};
