#ifndef ICI_CONF_H
#define ICI_CONF_H

/*
 * The following macros control the inclusion of various ICI intrinsic
 * functions. If the macro is defined the functions are NOT include
 * (hence the NO at the start of the name). By undef'ing the macro you
 * are stating that the functions should be included and compile (and
 * possibly even work) for the particular port of ICI.
 */
#undef  NOMATH          /* Trig and etc. */
#define NOWIN           /* Lib curses based screen handling. */
#define NODB            /* Simple ascii file based data base. */
#define NOTRACE         /* For debugging. */
#define NOWAITFOR       /* Requires select() or similar system primitive. */
#undef  NOSYSTEM        /* Command interpreter (shell) escape. */
#undef  NOPIPES         /* Requires popen(). */
#undef  NOSKT           /* BSD style network interface. */
#undef  NOSYSCALL       /* A few UNIX style system calls. */
#undef  NODIR           /* Directory reading function, dir(). */
#define NOPASSWD        /* UNIX password file access */
#undef  NODLOAD         /* Dynamic loading of native machine code modules. */
#define NOSTARTUPFILE   /* Parse a standard file of ICI code at init time. */
#undef  NODEBUGGING     /* Debugger interface and functions */
#undef  NOEVENTS        /* Event loop and associated processing. */
#define NOPROFILE       /* Profiler, see profile.c. */

/*
 * Mentioned in the version string.
 */
#define CONFIG_STR      "Microsoft Win32 platforms"

#define popen           _popen
#define pclose          _pclose
#define access          _access
#define ICI_PATH_SEP    ';'
#define ICI_DLL_EXT     ".dll"

/*
 * ICI_DLL is defined by dynamic link ICI libraries before they include
 * any ICI includes. This allows us to declare data items as imports into
 * the DLL. Otherwise they are not visible. See fwd.h.
 */
#ifdef  ICI_DLL
#define DLI   __declspec(dllimport)
#endif

#ifndef PREFIX
#define PREFIX "C:/ICI"
#endif

#ifndef _WIN32
#define _WIN32
#endif

int ici_get_last_win32_error(void);

#endif /*ICI_CONF_H*/
