#include "int.h"
#include "primes.h"

/*
 * Mark this and referenced unmarked objects, return memory costs.
 * See comments on t_mark() in object.h.
 */
static unsigned long
mark_int(object_t *o)
{
    o->o_flags |= O_MARK;
    return sizeof(int_t);
}

/*
 * Returns 0 if these objects are equal, else non-zero.
 * See the comments on t_cmp() in object.h.
 */
static int
cmp_int(object_t *o1, object_t *o2)
{
    return intof(o1)->i_value != intof(o2)->i_value;
}

/*
 * Return a hash sensitive to the value of the object.
 * See the comment on t_hash() in object.h
 */
static unsigned long
hash_int(object_t *o)
{
    /*
     * There are in-line versions of this in object.c and binop.h.
     */
    return (unsigned long)intof(o)->i_value * INT_PRIME;
}

/*
 * Return the int object with the value v. The returned object has had
 * its ref count incremented.
 */
int_t *
new_int(long v)
{
    register int_t	*i;

    /*
     * There is an in-line copy of this near the end of binop.h
     */
    if ((i = atom_int(v)) != NULL)
    {
	incref(i);
	return i;
    }
    if ((i = talloc(int_t)) == NULL)
	return NULL;
    objof(i)->o_type = &int_type;
    objof(i)->o_tcode = TC_INT;
    objof(i)->o_flags = 0;
    objof(i)->o_nrefs = 1;
    rego(i);
    i->i_value = v;
    return intof(atom(objof(i), 1));
}

type_t	int_type =
{
    mark_int,
    free_simple,
    hash_int,
    cmp_int,
    copy_simple,
    assign_simple,
    fetch_simple,
    "int"
};
