#include <fwd.h>
#ifndef NOSYSCALL

#include "func.h"
#include "op.h"
#include "exec.h"
#include "int.h"
#include "str.h"
#include "file.h"
#include "wrap.h"
#ifndef NODIR
#include "buf.h"
#include "re.h"
#endif /* NODIR */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>
#ifndef _WIN32
#include <unistd.h>
#endif
#if defined(sun) || defined(BSD4_4) || defined(linux)
#include <sys/wait.h>
#endif
#ifdef NeXT
#include <libc.h>
#endif

#ifdef SUNOS5
#include <signal.h>
#endif

#ifdef IRIX
#include <signal.h>
#include <sys/param.h>
#endif

#if defined(linux)
#include <sys/signal.h>
#endif

#ifndef NOPASSWD
#include <pwd.h>
#endif

#ifndef NODIR
#ifndef _WIN32
#include <sys/param.h>
#include <dirent.h>
#endif   /* _WIN32 */
#endif   /* NODIR */

#if defined(linux) && !defined(MAXPATHLEN)
#include <sys/param.h>
#endif

#ifdef  BSD4_4
#define SETPGRP_2_ARGS
#endif

#ifdef __BEOS__
#include <signal.h>
#endif

#ifdef _WIN32
#include <io.h>
#include <process.h>
#include <direct.h>

/*
 * Some macros for compatibility with Unix
 */

#define F_OK    0
#define W_OK    2
#define R_OK    4

#endif

/*
 * Create pre-defined variables to replace C's #define's.
 */
int
syscall_mkvars(struct_t *scp)
{
    int         i;

#define VALOF(x) { # x , x }
    static struct
    {
        char    *name;
        long    val;
    }
    var[] =
    {
        VALOF(O_RDONLY),
        VALOF(O_WRONLY),
        VALOF(O_RDWR),
        VALOF(O_APPEND),
        VALOF(O_CREAT),
        VALOF(O_TRUNC),
        VALOF(O_EXCL),
#ifdef O_BINARY         /* WIN32 has binary mode for open() */
        VALOF(O_BINARY),
#endif
#ifdef O_SYNC           /* WIN32 doesn't have O_SYNC */
        VALOF(O_SYNC),
#endif
#ifdef O_NDELAY         /* WIN32 doesn't have O_NDELAY */
        VALOF(O_NDELAY),
#endif
#ifdef O_NONBLOCK       /* NeXT doesn't have O_NONBLOCK */
        VALOF(O_NONBLOCK),
#endif
        VALOF(R_OK),
        VALOF(W_OK),
#ifdef X_OK             /* WIN32 doesn't have X_OK */
        VALOF(X_OK),
#endif
        VALOF(F_OK),
        VALOF(SEEK_SET),
        VALOF(SEEK_CUR),
        VALOF(SEEK_END),

#ifdef _WIN32
        VALOF(_P_WAIT),
        VALOF(_P_NOWAIT),
#endif

    };
#undef VALOF

    for (i = 0; i < sizeof var/sizeof var[0]; ++i)
        if (ici_cmkvar(scp, var[i].name, 'i', &var[i].val))
            return 1;
    return 0;
}

#if !defined(_WIN32) && !defined(__BEOS__)
/*
 * System calls.  We do all the easy ones because they cost so little,
 * most of the code being in the kernel.  And who can tell what people
 * will want anyway?
 */
#ifndef NeXT

extern int      access(), acct(), chdir(),
                chmod(), chown(), chroot(), close(),
#ifndef linux
                creat(),
#else
                __open(), /* creat() is a macro! defined in fcntl.h */
#endif
                dup(),
#if !defined(SUNOS5) && !defined(__hppa) && !defined(BSD4_4) && !defined(__BEOS__)
                fork(), getpid(), getpgrp(),
                getppid(),
#endif
#ifndef IRIX
                kill(), link(), mkdir(), mknod(),
                nice(), open(), pause(), rmdir(),
#endif
#if !defined(SUNOS5) && !defined(__hppa) && !defined(IRIX)
                setpgrp(),
#endif
                setuid(), setgid(), setgid(),

#if !defined(SUNOS5) && !defined(__hppa) && !defined(BSD4_4) && !defined(IRIX)
                sync(),
#endif

#if !defined(linux) && !defined(__hppa) && !defined(BSD4_4) && !defined(sun) && !defined(IRIX) && !defined(__BEOS__)
                umask(), alarm(), getuid(), geteuid(), getgid(), getegid(),
                lseek(), sleep(),
#endif

                ulimit(), unlink();

#if !defined(SUNOS5) && !defined(__hppa) && !defined(__USE_FIXED_PROTOTYPES__) && !defined(BSD4_4) && !defined(linux) && !defined(IRIX) && !defined(__BEOS__)
extern int      clock();
#endif

extern int      system(), lockf();

#endif

#ifdef sun
int rename();
int lstat();
void setpwent();
void endpwent();
#endif

#ifdef NeXT
int     acct();
int     getppid(void);
int     lockf();
#endif
#endif /* _WIN32 */

/*
 * Used as a common error return for system calls that fail. Sets the
 * global error string if the call fails otherwise returns the integer
 * result of the system call.
 */
static int
sys_ret(ret)
int     ret;
{
    if (ret < 0)
    {
        error = syserr();
        return 1;
    }
    return int_ret((long)ret);
}

/*
 * Used to call "simple" system calls. Simple calls are those that take
 * up to four integer parameters and return an integer.
 */
static int
sys_simple()
{
    long        av[4];

    if (ici_typecheck(CF_ARG2(), &av[0], &av[1], &av[2], &av[3]))
        return 1;
    return sys_ret((*CF_ARG1())(av[0], av[1], av[2], av[3]));
}

/*
 * String objects required by various calls (the struct object returned
 * by stat mostly).
 */
#define DEFSTRING(x)    static string_t *string_ ## x
#define GETSTRING(x)    if (need_string(&string_ ## x, #x)) return 1
#define DECREFSTRING(x) decref(string_ ## x)

DEFSTRING(dev);
DEFSTRING(ino);
DEFSTRING(mode);
DEFSTRING(nlink);
DEFSTRING(uid);
DEFSTRING(gid);
DEFSTRING(rdev);
DEFSTRING(size);
DEFSTRING(atime);
DEFSTRING(mtime);
DEFSTRING(ctime);
DEFSTRING(blksize);
DEFSTRING(blocks);

#ifndef NOPASSWD
DEFSTRING(name);
DEFSTRING(passwd);
/* DEFSTRING(uid) */    /* Already defined for stat() */
/* DEFSTRING(gid) */    /* Already defined for stat() */
DEFSTRING(gecos);
DEFSTRING(dir);
DEFSTRING(shell);
#endif

static int      need_strings = 1;

static int
wrapup_syscall()
{
DECREFSTRING(dev);
DECREFSTRING(ino);
DECREFSTRING(mode);
DECREFSTRING(nlink);
DECREFSTRING(uid);
DECREFSTRING(gid);
DECREFSTRING(rdev);
DECREFSTRING(size);
DECREFSTRING(atime);
DECREFSTRING(mtime);
DECREFSTRING(ctime);
DECREFSTRING(blksize);
DECREFSTRING(blocks);

#ifndef NOPASSWD
DECREFSTRING(name);
DECREFSTRING(passwd);
/* DECREFSTRING(uid) */    /* Already defined for stat() */
/* DECREFSTRING(gid) */    /* Already defined for stat() */
DECREFSTRING(gecos);
DECREFSTRING(dir);
DECREFSTRING(shell);
#endif

    return 0;
}

static int
get_strings(void)
{
    static wrap_t	wrap;

    GETSTRING(dev);
    GETSTRING(ino);
    GETSTRING(mode);
    GETSTRING(nlink);
    GETSTRING(uid);
    GETSTRING(gid);
    GETSTRING(rdev);
    GETSTRING(size);
    GETSTRING(atime);
    GETSTRING(mtime);
    GETSTRING(ctime);
    GETSTRING(blksize);
    GETSTRING(blocks);
#ifndef NOPASSWD
    GETSTRING(name);
    GETSTRING(passwd);
    /* GETSTRING(uid) */        /* Already defined for stat() */
    /* GETSTRING(gid) */        /* Already defined for stat() */
    GETSTRING(gecos);
    GETSTRING(dir);
    GETSTRING(shell);
#endif

    /* Ensure that strings are decref'd when ICI completes. */
    wrap.w_func = wrapup_syscall;
    wrap.w_next = wraps;
    wraps = &wrap;

    return need_strings = 0;
}

/*
 * open(pathname, flags [, mode])
 *
 * Do the open(2) system call with the specified flags and mode (only needed
 * if flags includes O_CREAT).
 *
 * Returns the file descriptor (an int).
 */
static int
f_open(void)
{
    char        *fname;
    long        omode;
    long        perms = -1; /* -1 means not passed */

    switch (NARGS())
    {
    case 2:
        if (ici_typecheck("si", &fname, &omode))
            return 1;
        break;
    case 3:
        if (ici_typecheck("sii", &fname, &omode, &perms))
            return 1;
        break;
    default:
        return ici_argcount(3);
    }
    if (omode & O_CREAT && perms == -1)
    {
        error = "permission bits not specified in open()";
        return 1;
    }
    return sys_ret(open(fname, omode, perms));
}

#ifndef _WIN32  /* WIN32 can't do fdopen() without lots of work */
/*
 * fdopen(int [, openmode])
 *
 * Open a file descriptor as an (ICI) file object. This is analogous
 * to fopen, sopen, mopen, popen, sktopen, etc... The openmode is a
 * string as accepted by fopen and defaults to "r" (read only).
 *
 * Returns a file object.
 */
static int
f_fdopen(void)
{
    long        fd;
    char        *mode;
    FILE        *stream;
    file_t      *f;

    switch (NARGS())
    {
    case 1:
        if (ici_typecheck("i", &fd))
            return 1;
        mode = "r";
        break;
    case 2:
        if (ici_typecheck("is", &fd, &mode))
            return 1;
        break;
    default:
        return ici_argcount(2);
    }
    if ((stream = fdopen(fd, mode)) == NULL)
    {
        error = "can't fdopen";
        return 1;
    }
    setvbuf(stream, NULL, _IOLBF, 0);
    if ((f = new_file((char *)stream, &stdio_ftype, NULL)) == NULL)
    {
        fclose(stream);
        return 1;
    }
    return ici_ret_with_decref(objof(f));
}
#endif /* _WIN32 */

/*
 * stat(string)
 * stat(int)
 * stat(file)
 *
 * Perform the stat(2) or fstat(2) system call. The file to be stat'd can
 * be passed as a pathname (string), ICI file object (returned by fopen)
 * or file descriptor (int).
 *
 * Returns a struct containing integers with the following keys,
 *      atime           blksize         blocks          ctime
 *      dev             gid             ino             mode
 *      mtime           nlink           rdev            size
 *      uid
 *
 * These correspond to the fields of the same names (with a st_ prefix)
 * in the struct stat returned by the system call.
 */
static int
f_stat(void)
{
    object_t    *o;
    struct stat statb;
    int         rc;
    struct_t    *s;

    if (NARGS() != 1)
        return ici_argcount(1);
    o = ARG(0);
    if (isint(o))
        rc = fstat(intof(o)->i_value, &statb);
    else if (isstring(o))
        rc = stat(stringof(o)->s_chars, &statb);
    else if (isfile(o) && fileof(o)->f_type == &stdio_ftype)
        rc = fstat(fileno((FILE *)fileof(o)->f_file), &statb);
    else
        return ici_argerror(0);
    if (rc == -1)
        return sys_ret(rc);
    if ((s = new_struct()) == NULL || (need_strings && get_strings()))
        return 1;
#define SETFIELD(x)                                             \
    if ((o = objof(new_int(statb.st_ ## x))) == NULL)           \
        goto fail;                                              \
    else if (assign(s, string_ ## x, o))                        \
    {                                                           \
        decref(o);                                              \
        goto fail;                                              \
    }                                                           \
    else                                                        \
        decref(o)

    SETFIELD(dev);
    SETFIELD(ino);
    SETFIELD(mode);
    SETFIELD(nlink);
    SETFIELD(uid);
    SETFIELD(gid);
    SETFIELD(rdev);
    SETFIELD(size);
    SETFIELD(atime);
    SETFIELD(mtime);
    SETFIELD(ctime);
#ifndef _WIN32
    SETFIELD(blksize);
#endif /* _WIN32 */

#if !defined(__BEOS__) && !defined(_WIN32)
    SETFIELD(blocks);
#endif /* __BEOS__ , _WIN32 */


#undef SETFIELD

    return ici_ret_with_decref(objof(s));

fail:
    decref(s);
    return 1;
}

/*
 * ctime(int)
 *
 * Convert a time value to a string. Takes a count of the number of
 * seconds past the epoch and returns a string (see ctime(3)).
 */
static int
f_ctime(void)
{
    long        timev;
    string_t    *s;

    if (ici_typecheck("i", &timev) || (s = new_cname(ctime(&timev))) == NULL)
        return 1;
    return ici_ret_with_decref(objof(s));
}

/*
 * time()
 *
 * time() returns the time since 00:00:00 GMT,  Jan.  1,  1970 (the epoch)
 * measured in seconds.
 */
static int
f_time(void)
{
    return int_ret(time(NULL));
}

/*
 * access(pathname [,flags])
 *
 * Determine accessibility of a file. If flags is not passed it defaults
 * to F_OK.
 *
 */
static int
f_access(void)
{
    char        *fname;
    int         bits = F_OK;

    switch (NARGS())
    {
    case 1:
        if (ici_typecheck("s", &fname))
            return 1;
        break;
    case 2:
        if (ici_typecheck("si", &fname, &bits))
            return 1;
        break;
    default:
        return ici_argcount(2);
    }
    return sys_ret(access(fname, bits));
}

#ifndef _WIN32
/*
 * pipe()
 *
 * Create a pipe and return an array of two file descriptors.
 */
static int
f_pipe(void)
{
    int         pfd[2];
    array_t     *a;
    int_t       *fd;

    if ((a = new_array()) == NULL)
        return 1;
    if (pipe(pfd) == -1)
    {
        decref(a);
        return sys_ret(-1);
    }
    if ((fd = new_int(pfd[0])) == NULL)
        goto fail;
    decref(fd);
    if (assign(a, o_zero, fd))
        goto fail;
    if ((fd = new_int(pfd[1])) == NULL)
        goto fail;
    decref(fd);
    if (assign(a, o_one, fd))
        goto fail;
    return ici_ret_with_decref(objof(a));

fail:
    decref(a);
    close(pfd[0]);
    close(pfd[1]);
    return 1;
}
#endif  /* #ifndef _WIN32 */

/*
 * creat(pathname, mode)
 *
 * Create the named file with the given permissions.
 */
static int
f_creat(void)
{
    char        *fname;
    long        perms;
    int         fd;

    if (ici_typecheck("si", &fname, &perms))
        return 1;
    if ((fd = creat(fname, perms)) == -1)
    {
        error = syserr();
        return 1;
    }
    return int_ret(fd);
}

/*
 * dup(int [, int])
 *
 * Perform a dup(2) or dup2(2) system call depending upon args.
 */
static int
f_dup(void)
{
    long        fd1;
    long        fd2;

    switch (NARGS())
    {
    case 1:
        if (ici_typecheck("i", &fd1))
            return 1;
        fd2 = -1;
        break;
    case 2:
        if (ici_typecheck("ii", &fd1, &fd2))
            return 1;
        break;
    default:
        return ici_argcount(2);
    }
    if (fd2 == -1)
    {
        fd2 = dup(fd1);
        return fd2 < 0 ? sys_ret(fd2) : int_ret(fd2);
    }
    if (dup2(fd1, fd2) < 0)
        return sys_ret(-1);
    return int_ret(fd2);
}

/*
 * exec(pathname, array)
 * exec(pathname, arg0, arg1...)
 *
 * Execute a program. The first form uses an array to store the
 * argyuments passed to the program, the second uses a variable
 * number of arguments.
 */
static int
f_exec(void)
{
    char        *sargv[16];
    char        **argv;
    int         maxargv;
    int         n;
    object_t    **o;
    char        *path;
    int         argc;

#define ADDARG(X)                                                       \
    {                                                                   \
        if (argc >= 16)                                                 \
        {                                                               \
            if (argc >= maxargv)                                        \
            {                                                           \
                char    **newp;                                         \
                int      i;                                             \
                maxargv += argc;                                        \
                if ((newp = ici_alloc(maxargv * sizeof (char *))) == NULL)      \
                {                                                       \
                    if (argv != sargv)                                  \
                        ici_free(argv);                                 \
                    return 1;                                           \
                }                                                       \
                for (i = 0; i < argc; ++i)                              \
                    newp[i] = argv[i];                                  \
                if (argv != sargv)                                      \
                    ici_free(argv);                                     \
                argv = newp;                                            \
            }                                                           \
        }                                                               \
        argv[argc++] = (X);                                             \
    }

    if ((n = NARGS()) < 2)
        return ici_argcount(2);
    if (!isstring(*(o = ARGS())))
        return ici_argerror(0);
    path = stringof(*o)->s_chars;
    --o;
    argc = 0;
    argv = sargv;
    maxargv = 16;
    if (n > 2 || isstring(*o))
    {
        while (--n > 0)
        {
            ADDARG(stringof(*o)->s_chars);
            --o;
        }
    }
    else if (!isarray(*o))
        return ici_argerror(0);
    else
    {
        object_t **p;
        for (p = arrayof(*o)->a_base; p < arrayof(*o)->a_top; ++p)
            if (isstring(*p))
                ADDARG(stringof(*p)->s_chars);
    }
    ADDARG(NULL);
    n = (*CF_ARG1())(path, argv);
    if (argv != sargv)
        ici_free(argv);
    return sys_ret(n);
}

#ifdef _WIN32
/*
 * spawn([mode,] pathname, array)
 * spawn([mode,] pathname, arg0, arg1...)
 * spawnp([mode,] pathname, array)
 * spawnp([mode,] pathname, arg0, arg1...)
 *
 * Execute a program. The first form uses an array to store the
 * argyuments passed to the program, the second uses a variable
 * number of arguments.
 */
static int
f_spawn(void)
{
    char        *sargv[16];
    char        **argv;
    int         maxargv;
    int         n;
    object_t    **o;
    char        *path;
    int         argc;
    int         mode = _P_NOWAIT;

#define ADDARG(X)                                                       \
    {                                                                   \
        if (argc >= 16)                                                 \
        {                                                               \
            if (argc >= maxargv)                                        \
            {                                                           \
                char    **newp;                                         \
                int      i;                                             \
                maxargv += argc;                                        \
                if ((newp = ici_alloc(maxargv * sizeof (char *))) == NULL)      \
                {                                                       \
                    if (argv != sargv)                                  \
                        ici_free(argv);                                 \
                    return 1;                                           \
                }                                                       \
                for (i = 0; i < argc; ++i)                              \
                    newp[i] = argv[i];                                  \
                if (argv != sargv)                                      \
                    ici_free(argv);                                     \
                argv = newp;                                            \
            }                                                           \
        }                                                               \
        argv[argc++] = (X);                                             \
    }

    if ((n = NARGS()) < 2)
        return ici_argcount(2);
    o = ARGS();
    if (isint(*o))
    {
        mode = intof(*o)->i_value;
        --o;
        if (--n < 2)
            return ici_argcount(2);
        if (!isstring(*o))
            return ici_argerror(1);
    }
    else if (!isstring(*o))
        return ici_argerror(0);
    path = stringof(*o)->s_chars;
    --o;
    argc = 0;
    argv = sargv;
    maxargv = 16;
    if (n > 2 || isstring(*o))
    {
        while (--n > 0)
        {
            ADDARG(stringof(*o)->s_chars);
            --o;
        }
    }
    else if (!isarray(*o))
        return ici_argerror(0);
    else
    {
        object_t **p;
        for (p = arrayof(*o)->a_base; p < arrayof(*o)->a_top; ++p)
            if (isstring(*p))
                ADDARG(stringof(*p)->s_chars);
    }
    ADDARG(NULL);
    n = (*CF_ARG1())(mode, path, argv);
    if (argv != sargv)
        ici_free(argv);
    if (n == -1)
    {
        error = syserr();
        return 1;
    }
    return int_ret(n);
}
#endif

/*
 * lseek(int, int [, int])
 *
 * Set a file descriptor's I/O position. If 'whence' not passed
 * it defaults to SEEK_SET.
 */
static int
f_lseek(void)
{
    int_t       *fd;
    long        ofs;
    long        whence = SEEK_SET;

    switch (NARGS())
    {
    case 2:
        if (ici_typecheck("ii", &fd, &ofs))
            return 1;
        break;
    case 3:
        if (ici_typecheck("iii", &fd, &ofs, &whence))
            return 1;
    default:
        return ici_argcount(3);
    }
    return sys_ret(lseek(fd->i_value, ofs, whence));
}

#ifndef _WIN32
/*
 * wait()
 *
 * Wait for a child process to exit and return the exit status.
 */
static int
f_wait(void)
{
#if NeXT
    union wait  status;
#else
    int status;
#endif
    if (wait(&status) < 0)
        return sys_ret(-1);
#if NeXT
    return int_ret(status.w_retcode);
#else
    return int_ret(status);
#endif
}
#endif /* _WIN32 */

#ifndef NODIR

#ifdef _WIN32
/*
 * Emulate opendir/readdir/et al under WIN32 environments via findfirst/
 * findnext. Only what f_dir() needs has been emulated (which is to say,
 * not much).
 */

#define MAXPATHLEN      _MAX_PATH

struct dirent
{
    char        *d_name;
};

typedef struct DIR
{
    long                handle;
    struct _finddata_t  finddata;
    int                 needfindnext;
    struct dirent       dirent;
}
DIR;

static DIR *
opendir(const char *path)
{
    DIR         *dir;
    char        fspec[_MAX_PATH+1];

    if (strlen(path) > (_MAX_PATH - 4))
        return NULL;
    sprintf(fspec, "%s/*.*", path);
    if ((dir = talloc(DIR)) != NULL)
    {
        if ((dir->handle = _findfirst(fspec, &dir->finddata)) == -1)
        {
            ici_free(dir);
            return NULL;
        }
        dir->needfindnext = 0;
    }
    return dir;
}

static struct dirent *
readdir(DIR *dir)
{
    if (dir->needfindnext && _findnext(dir->handle, &dir->finddata) != 0)
            return NULL;
    dir->dirent.d_name = dir->finddata.name;
    dir->needfindnext = 1;
    return &dir->dirent;
}

static void
closedir(DIR *dir)
{
    _findclose(dir->handle);
    ici_free(dir);
}

#define S_ISREG(m)      (((m) & _S_IFMT) == _S_IFREG)
#define S_ISDIR(m)      (((m) & _S_IFMT) == _S_IFDIR)

#endif

/*
 * array = dir([path] [, regexp] [, format])
 *
 * Read directory named in path (a string, defaulting to ".", the current
 * working directory) and return the entries that match the pattern (or
 * all names if no pattern passed). The format string identifies what
 * sort of entries should be returned. If the format string is passed
 * then a path MUST be passed (to avoid any ambiguity) but path may be
 * NULL meaning the current working directory (same as "."). The format
 * string uses the following characters,
 *
 *      f       return file names
 *      d       return directory names
 *      a       return all names (which includes things other than
 *              files and directories, e.g., hidden or special files
 *
 * The default format specifier is "f".
 */
static int
f_dir(void)
{
    char                *path   = ".";
    char                *format = "f";
    regexp_t            *regexp = NULL;
    object_t            *o;
    array_t             *a;
    DIR                 *dir;
    struct dirent       *dirent;
    int                 fmt;
    string_t            *s;

    switch (NARGS())
    {
    case 0:
        break;

    case 1:
        o = ARG(0);
        if (isstring(o))
            path = stringof(o)->s_chars;
        else if (isnull(o))
            ;   /* leave path as is */
        else if (isregexp(o))
            regexp = regexpof(o);
        else
            return ici_argerror(0);
        break;

    case 2:
        o = ARG(0);
        if (isstring(o))
            path = stringof(o)->s_chars;
        else if (isnull(o))
            ;   /* leave path as is */
        else if (isregexp(o))
            regexp = regexpof(o);
        else
            return ici_argerror(0);
        o = ARG(1);
        if (isregexp(o))
        {
            if (regexp != NULL)
                return ici_argerror(1);
            regexp = regexpof(o);
        }
        else if (isstring(o))
            format = stringof(o)->s_chars;
        else
            return ici_argerror(1);
        break;

    case 3:
        o = ARG(0);
        if (isstring(o))
            path = stringof(o)->s_chars;
        else if (isnull(o))
            ;   /* leave path as is */
        else
            return ici_argerror(0);
        o = ARG(1);
        if (!isregexp(o))
            return ici_argerror(1);
        regexp = regexpof(o);
        o = ARG(2);
        if (!isstring(o))
            return ici_argerror(2);
        format = stringof(o)->s_chars;
        break;

    default:
        return ici_argcount(3);
    }

    if (*path == '\0')
        path = ".";

#define FILES   1
#define DIRS    2
#define OTHERS  4

    for (fmt = 0; *format != '\0'; ++format)
    {
        switch (*format)
        {
        case 'f':
            fmt |= FILES;
            break;

        case 'd':
            fmt |= DIRS;
            break;

        case 'a':
            fmt |= OTHERS | DIRS | FILES;
            break;

        default:
            error = "bad directory format specifier";
            return 1;
        }
    }
    if ((a = new_array()) == NULL)
        return 1;
    if ((dir = opendir(path)) == NULL)
    {
        if (chkbuf(strlen(path) + 32))
            error = syserr();
        else
        {
            sprintf(buf, "%s: %s", path, syserr());
            error = buf;
        }
        goto fail;
    }
    while ((dirent = readdir(dir)) != NULL)
    {
        struct stat     statbuf;
        char            abspath[MAXPATHLEN+1];

        if
        (
            regexp != NULL
            &&
            pcre_exec
            (
                regexp->r_re,
                regexp->r_rex,
                dirent->d_name,
                strlen(dirent->d_name),
#if defined(PCRE_MINOR) && (PCRE_MINOR >= 8)
		0,
#endif
                0,
                re_bra,
                nels(re_bra)
            )
            < 0
        )
            continue;
        sprintf(abspath, "%s/%s", path, dirent->d_name);
#ifndef _WIN32
        if (lstat(abspath, &statbuf) == -1)
        {
            if (chkbuf(strlen(abspath) + 32))
                error = syserr();
            else
            {
                sprintf(buf, "%s: %s", abspath, syserr());
                error = buf;
            }
            closedir(dir);
            goto fail;
        }
        if (S_ISLNK(statbuf.st_mode) && stat(abspath, &statbuf) == -1)
            continue;
#else
        if (stat(abspath, &statbuf) == -1)
            continue;
#endif
        if
        (
            (S_ISREG(statbuf.st_mode) && fmt & FILES)
            ||
            (S_ISDIR(statbuf.st_mode) && fmt & DIRS)
            ||
            fmt & OTHERS
        )
        {
            if
            (
                (s = new_cname(dirent->d_name)) == NULL
                ||
                pushcheck(a, 1)
            )
            {
                if (s != NULL)
                    decref(s);
                closedir(dir);
                goto fail;
            }
            *a->a_top++ = objof(s);
            decref(s);
        }
    }
    closedir(dir);
    return ici_ret_with_decref(objof(a));

#undef  FILES
#undef  DIRS
#undef  OTHERS

fail:
    decref(a);
    return 1;
}
#endif  /* NODIR */

/*
 * string = getcwd()
 */
static int
f_getcwd(void)
{
    char        buf[MAXPATHLEN+1];

    if (getcwd(buf, sizeof buf) == NULL)
        return sys_ret(-1);
    return str_ret(buf);
}

#ifndef NOPASSWD
/*
 * array = passwd()
 * struct = passwd(string)
 * struct = passwd(int)
 *
 * Return an array of all password file entries or a single entry if given a user name
 * or user id. In the later case return NULL if no entry exists for that name/uid.
 *
 * Each password file entry is returned as a struct with the following keys,
 *
 *      name
 *      passwd
 *      uid
 *      gid
 *      gecos
 *      dir
 *      shell
 */
static struct_t *password_struct(struct passwd *);

static int
f_passwd(void)
{
    struct passwd       *pwent;
    array_t             *a;

    switch (NARGS())
    {
    case 0:
        break;

    case 1:
        if (isint(ARG(0)))
            pwent = getpwuid((uid_t)intof(ARG(0))->i_value);
        else if (isstring(ARG(0)))
            pwent = getpwnam(stringof(ARG(0))->s_chars);
        else
            return ici_argerror(0);
        if (pwent == NULL)
            error = "no such user";
        return ici_ret_with_decref(objof(password_struct(pwent)));

    default:
        return ici_argcount(1);
    }

    if ((a = new_array()) == NULL)
        return 1;
    setpwent();
    while ((pwent = getpwent()) != NULL)
    {
        struct_t        *s;

        if (pushcheck(a, 1) || (s = password_struct(pwent)) == NULL)
        {
            decref(a);
            return 1;
        }
        *a->a_top++ = objof(s);
    }
    endpwent();
    return ici_ret_with_decref(objof(a));
}

static struct_t *
password_struct(struct passwd *pwent)
{
    struct_t    *d;
    object_t    *o;

    if (pwent == NULL)
        return NULL;
    if (need_strings && get_strings())
        return NULL;
    if ((d = new_struct()) != NULL)
    {

#define SET_INT_FIELD(x)                                        \
    if ((o = objof(new_int(pwent->pw_ ## x))) == NULL)          \
        goto fail;                                              \
    else if (assign(d, string_ ## x, o))                        \
    {                                                           \
        decref(o);                                              \
        goto fail;                                              \
    }                                                           \
    else                                                        \
        decref(o)

#define SET_STR_FIELD(x)                                        \
    if ((o = objof(new_cname(pwent->pw_ ## x))) == NULL)        \
        goto fail;                                              \
    else if (assign(d, string_ ## x, o))                        \
    {                                                           \
        decref(o);                                              \
        goto fail;                                              \
    }                                                           \
    else                                                        \
        decref(o)

        SET_STR_FIELD(name);
        SET_STR_FIELD(passwd);
        SET_INT_FIELD(uid);
        SET_INT_FIELD(gid);
        SET_STR_FIELD(gecos);
        SET_STR_FIELD(dir);
        SET_STR_FIELD(shell);

#undef SET_STR_FIELD
#undef SET_INT_FIELD

    }
    return d;

fail:
    decref(d);
    return NULL;
}
#endif

#ifdef SETPGRP_2_ARGS
static int
f_setpgrp(void)
{
    long        pid, pgrp;

    switch (NARGS())
    {
    case 1:
        pid = 0;
        if (ici_typecheck("i", &pgrp))
            return 1;
        break;

    default:
        if (ici_typecheck("ii", &pid, &pgrp))
            return 1;
        break;
    }
    return sys_ret(setpgrp((pid_t)pid, (pid_t)pgrp));
}
#endif

cfunc_t syscall_cfuncs[] =
{
    /*
     * Things that don't use sys_simple
     */
    {CF_OBJ,    "access",       f_access},
    {CF_OBJ,    "creat",        f_creat},
    {CF_OBJ,    "dup",          f_dup},
    {CF_OBJ,    "exec",         f_exec,         execv},
    {CF_OBJ,    "execp",        f_exec,         execvp},
#ifdef _WIN32
    {CF_OBJ,    "spawn",        f_spawn,        spawnv},
    {CF_OBJ,    "spawnp",       f_spawn,        spawnvp},
#endif
    {CF_OBJ,    "lseek",        f_lseek},
    {CF_OBJ,    "open",         f_open},
#ifndef _WIN32
    {CF_OBJ,    "pipe",         f_pipe},
#endif
    {CF_OBJ,    "stat",         f_stat},
#ifndef _WIN32
    {CF_OBJ,    "wait",         f_wait},
#endif
    {CF_OBJ,    "ctime",        f_ctime},
    {CF_OBJ,    "time",         f_time},
#ifndef _WIN32
    {CF_OBJ,    "fdopen",       f_fdopen},
#endif
    {CF_OBJ,    "getcwd",       f_getcwd},

    /*
     * Many system calls are "simple", i.e., they take a restricted
     * number of parameters and return an int, sys_simple is provided
     * to call these and the SIMPLE macro is used to define them in
     * our cfunc_t table.
     */

#define SIMPLE(x)       sys_simple, ((int (*)()) x)

#ifndef _WIN32
    /*
     * "Simple" system calls
     */
    {CF_OBJ,    "alarm",        SIMPLE(alarm),          "i"},
#ifndef __BEOS__
    {CF_OBJ,    "acct",         SIMPLE(acct),           "s"},
#endif

#endif /* _WIN32 */

    {CF_OBJ,    "chdir",        SIMPLE(chdir),          "s"},
#ifndef _WIN32
    {CF_OBJ,    "chmod",        SIMPLE(chmod),          "si"},
    {CF_OBJ,    "chown",        SIMPLE(chown),          "sii"},
#ifndef __BEOS__
    {CF_OBJ,    "chroot",       SIMPLE(chroot),         "s"},
#endif /* __BEOS__ */
#endif /* _WIN32 */

    {CF_OBJ,    "_close",       SIMPLE(close),          "i"},
    /*### fcntl */
    {CF_OBJ,    "_exit",        SIMPLE(exit),           "i"},
#ifndef _WIN32
    {CF_OBJ,    "fork",         SIMPLE(fork),           ""},
    {CF_OBJ,    "getpid",       SIMPLE(getpid),         ""},
    {CF_OBJ,    "getpgrp",      SIMPLE(getpgrp),        ""},
    {CF_OBJ,    "getppid",      SIMPLE(getppid),        ""},
    {CF_OBJ,    "getuid",       SIMPLE(getuid),         ""},
    {CF_OBJ,    "geteuid",      SIMPLE(geteuid),        ""},
    {CF_OBJ,    "getgid",       SIMPLE(getgid),         ""},
    {CF_OBJ,    "getegid",      SIMPLE(getegid),        ""},
    {CF_OBJ,    "kill",         SIMPLE(kill),           "ii"},
    {CF_OBJ,    "link",         SIMPLE(link),           "ss"},
#endif
    {CF_OBJ,    "mkdir",        SIMPLE(mkdir),          "si"},
#ifndef _WIN32
    {CF_OBJ,    "mknod",        SIMPLE(mknod),          "sii"},
#ifndef __BEOS__
    {CF_OBJ,    "nice",         SIMPLE(nice),           "i"},
#endif
    {CF_OBJ,    "pause",        SIMPLE(pause),          ""},
    /* ### poll */
    /* ### read */
#endif
    {CF_OBJ,    "rmdir",        SIMPLE(rmdir),          "s"},
#ifndef _WIN32
#ifdef SETPGRP_2_ARGS
    {CF_OBJ,    "setpgrp",      f_setpgrp},
#elif !defined(__BEOS__)
    {CF_OBJ,    "setpgrp",      SIMPLE(setpgrp),        ""},
#endif
    {CF_OBJ,    "setuid",       SIMPLE(setuid),         "i"},
    {CF_OBJ,    "setgid",       SIMPLE(setgid),         "i"},
    {CF_OBJ,    "signal",       SIMPLE(signal),         "ii"},
    {CF_OBJ,    "sync",         SIMPLE(sync),           ""},
    /* ### times */
#if !defined(NeXT) && !defined(BSD4_4) && !defined(__BEOS__)
    {CF_OBJ,    "ulimit",       SIMPLE(ulimit),         "ii"},
#endif
    {CF_OBJ,    "umask",        SIMPLE(umask),          "i"},
    /* ### uname */
#endif
    /* ### unlink() should go as remove() is more portable */
    {CF_OBJ,    "unlink",       SIMPLE(unlink),         "s"},
    /* ### utime */
    /* ### write */

#ifndef _WIN32
    /*
     * Non system calls, but they act so similar.
     */
    {CF_OBJ,    "clock",        SIMPLE(clock),          ""},
#ifdef NOTDEF
    /* Already f_system in clib.c! */
    {CF_OBJ,    "system",       SIMPLE(system),         "s"},
#endif

#if !defined(linux) && !defined(BSD4_4) && !defined(__BEOS__)
    {CF_OBJ,    "lockf",        SIMPLE(lockf),          "iii"},
#endif /* linux */
    {CF_OBJ,    "sleep",        SIMPLE(sleep),          "i"},
#endif /* _WIN32 */

#ifndef NODIR
    {CF_OBJ,    "dir",          f_dir},
#endif

#ifndef NOPASSWD
    {CF_OBJ,    "passwd",       f_passwd},
#endif

    {CF_OBJ,    "rename",       SIMPLE(rename),         "ss"},

    {CF_OBJ}
};
#endif /* NOSYSCALL */
