#include "buf.h"
#include "exec.h"
#include "wrap.h"


extern void	drop_fast_free_lists(void);


/*
 * Shut down the interpreter and clean up any allocations.
 *
 * This function is the reverse of ici_init(), and a superset of wrapup().
 */
void
ici_uninit(void)
{
    /* This catches the case where ici_uninit() is called without ici_init
     * ever being called. */
    if (o_zero == NULL)
        return;

    /* Clean up anything registered by modules that are only optionally
     * compiled in. */
    wrapup();

    /* Clean up ICI variables used by various bits of ICI. */
    decref(o_zero);
    o_zero = NULL;
    decref(o_one);
    o_one = NULL;

    /* Call uninitialisation functions for compulsory bits of ICI. */
    uninit_parse();
    uninit_compile();
    uninit_cfunc();
    uninit_func();

    /* Ensure that the stacks don't stop anything being collected. */
    v_top = vs->a_base;
    o_top = os->a_base;
    x_top = xs->a_base;

    /* The first call moves them to the fast free list and the second actually
     * cleans up and free()s their memory. */
    ici_reclaim();
    ici_reclaim();

    /* Delete the arrays allocated in the operator, execution and variable
     * stacks (we can't just delete them normally because they're statically
     * defined in exec.c. */
    if (arrayof(os)->a_base != NULL)
    {
        ici_free(arrayof(os)->a_base);
        arrayof(os)->a_base = NULL;
        ex.x_os.a_base = ex.x_os.a_limit = o_top = NULL;
    }
    if (arrayof(xs)->a_base != NULL)
    {
        ici_free(arrayof(xs)->a_base);
        arrayof(xs)->a_base = NULL;
        ex.x_xs.a_base = ex.x_xs.a_limit = x_top = NULL;
    }
    if (arrayof(vs)->a_base != NULL)
    {
        ici_free(arrayof(vs)->a_base);
        arrayof(vs)->a_base = NULL;
        ex.x_vs.a_base = ex.x_vs.a_limit = v_top = NULL;
    }

    /* Free the general purpose buffer. */
    ici_free(ici_buf);
    ici_buf = NULL;
    ici_bufz = 0;

    /* Destroy the now empty tables. */
    ici_free(atoms);
    atoms = NULL;
    ici_free(objs);
    objs = NULL;

    /* Cause the final freeing of any freed since the last ici_reclaim. */
    drop_fast_free_lists();
}
