#include <fwd.h>

#ifndef	NOWIN
/*
 * Text based screen windows manager.
 */
#include "exec.h"
#include "op.h"
#include "str.h"
#include "win.h"
#include "acs.h"
#include "buf.h"
#include "func.h"
#include "int.h"
#include "wrap.h"
#include <fcntl.h>

int		tputs();

int		ti_init();
int		ti_canon();
void		ti_nocanon();
char		*ti_getchar();
static char	*w_getchar();

#ifdef	TERMINFO
#define	tgoto(x, c, l)	tparm(x, l, c)
#endif

static win_t	*win_bottom;
static win_t	*win_top;
static win_t	*win_cursor;		/* Which window has the cursor. */
static int	win_isamess	= 1;	/* Dirt on wantscreen. */
static int	win_atline;		/* Current cursor position. */
static int	win_atcol;		/* Ditto. */
static int	win_lost	= 1;	/* Don't know cursor position. */
static win_t	*win_onscreen;
static win_t	*win_wantscreen;
static win_t	*mesg_window;
static int	mesg_transient;
static int	win_open;
static int	inacs;
static char	w_ungotchar[16];
static int	w_ungot;
static int	w_suspended;

/*
 * Terminal information.
 */
char	*ti_clear;	/* Clear screen. */
char	*ti_cup;	/* Cursor position. */
int	ti_ncols;
int	ti_nlines;
int	ti_automargins;
char	*ti_smacs;	/* Start mode, alternate character set. */
char	*ti_rmacs;	/* End mode, alternate character set. */
int	ti_gotocost;	/* Typical no. of chars in a goto seq. */
/*
 * Alternate character set maps.  These must be in sync with the #defines
 * in acs.h and the vt100 map in ti.c.
 */
char	ti_defmap[]	= " ><v##^+:  '##'..`+--_++++|*";
char	ti_acsmap[sizeof ti_defmap];

#ifndef	SMALL
#define	w_ptr(w,l,c)	((w)->w_image + ((c) + (w)->w_ncols * (l)))
#else
/*
 * Return pointer to character at given line and column in the window.
 * Caller's responsibility to ensure that pointer does not exceed bounds.
 */
char *
w_ptr(w, l, c)
register win_t	*w;
register int	l;
register int	c;
{
    return w->w_image + (c + w->w_ncols * l);
}
#endif

#ifndef	MSDOS

extern char	*tgoto();

#define	screen_clear()		tputs(ti_clear, ti_nlines, ti_putc)
#define	start_acs()		tputs(ti_smacs, 1, ti_putc)
#define	end_acs()		tputs(ti_rmacs, 1, ti_putc)
#define	screen_goto(line, col) 	tputs(tgoto(ti_cup, col, line), 1, ti_putc)

#ifndef TI_PUTC_ARG
#define	TI_PUTC_ARG int
#endif

int
ti_putc(TI_PUTC_ARG c)
{
    putchar(c);
    return 0;
}

#else

#include <dos.h>

#define	start_acs()		0
#define	end_acs()		0

static union REGS	inregs;
static union REGS	outregs;

static void
screen_clear()
{
    register int	line;
    register int	col;

    memset(&inregs, 0, sizeof inregs);
    for (line = 0; line < ti_nlines; ++line)
    {
	memset(&inregs, 0, sizeof inregs);
	inregs.h.ah = 2;	/* Set cursor position. */
	inregs.h.dl = 0;	/* Column 0. */
	inregs.h.dh = line;
	inregs.h.bh = 0;	/* Page no. for write character. */
	int86(0x10, &inregs, &outregs);	/* BIOS video function. */

	memset(&inregs, 0, sizeof inregs);
	inregs.h.ah = 9;	/* Write attribute/character at cursor pos. */
	inregs.h.al = ' ';	/* Character to write. */
	inregs.h.bl = 7;	/* Attributes. */
	inregs.h.bh = 0;	/* Page no. for write character. */
	inregs.x.cx = ti_ncols;	/* Repetition factor for write character. */
	int86(0x10, &inregs, &outregs); /* BIOS video function. */
    }
}

static void
screen_goto(line, col)
register int	line;
register int	col;
{
    memset(&inregs, 0, sizeof inregs);
    inregs.h.ah = 2;		/* Set cursor position. */
    inregs.h.dh = line;
    inregs.h.dl = col;
    inregs.h.bh = 0;
    int86(0x10, &inregs, &outregs); /* BIOS video function. */
}

static int
ti_putc(int c)
{
    memset(&inregs, 0, sizeof inregs);
    inregs.h.ah = 9;		/* Write attribute/character at cursor pos. */
    inregs.h.al = c;		/* Character to be written. */
    inregs.h.bl = 7;		/* Attributes to be written. */
    inregs.h.bh = 0;		/* Page no. for write character. */
    inregs.x.cx = 1;		/* Repetition factor. */
    int86(0x10, &inregs, &outregs); /* BIOS video function. */
    return 0;
}

#endif

void
w_isdirty(w, line)
register win_t	*w;
register int	line;
{
    if ((w->w_flags & W_ISDIRTY) == 0)
    {
	w->w_mindirt = line;
	w->w_maxdirt = line;
	w->w_flags |= W_ISDIRTY;
    }
    else
    {
	if (w->w_mindirt > line)
	    w->w_mindirt = line;
	if (w->w_maxdirt < line)
	    w->w_maxdirt = line;
    }
}

static void w_popall();
static void w_refresh();
static void win_putc();

int
w_finish()
{
    if (win_open)
    {
	win_putc(' '); /* Out of any graphics char set. */
	w_popall();
	w_refresh();
	win_open = 0;
    }
    return 0;
}

/*
 * Can fail, returns 1, usual conventions.
 */
static int
w_init()
{
    register win_t	*w;
    char		*e;
    static wrap_t	wrap;

    if (win_open)
	return 0;
#ifndef	MSDOS
    if (ti_init())
	return 1;
    fflush(stdout);
    setvbuf(stdout, NULL, _IOFBF, BUFSIZ);
#else
    ti_ncols = 80;
    ti_nlines = 25;
    ti_automargins = 0;
    ti_smacs = "";
    ti_rmacs = "";
    ti_gotocost = 0;
    memcpy
    (
	ti_acsmap,
	"X\32\33\23\333\17\30\4\262XX\370\361\261\331\277\332\300\305\304\304_\303\264\301\302\263\7",
	sizeof ti_acsmap
    );
#endif

    if ((e = (char *)getenv("LINES")) != NULL)
	ti_nlines = ici_strtol(e, NULL, 10);
    if ((e = (char *)getenv("COLS")) != NULL)
	ti_ncols = ici_strtol(e, NULL, 10);

    if ((w = talloc(win_t)) == NULL)
	return 1;
    if ((w->w_image = (char *)ici_alloc(ti_nlines * ti_ncols)) == NULL)
    {
	ici_free((char *)w);
	return 1;
    }
    win_onscreen = w;
    w->w_cline = 0;
    w->w_ccol = 0;
    w->w_nlines = ti_nlines;
    w->w_ncols = ti_ncols;

    if ((w = talloc(win_t)) == NULL)
    {
	ici_free(win_onscreen->w_image);
	ici_free((char *)win_onscreen);
	return 1;
    }
    if ((w->w_image = (char *)ici_alloc(ti_nlines * ti_ncols)) == NULL)
    {
	ici_free((char *)w);
	ici_free(win_onscreen->w_image);
	ici_free((char *)win_onscreen);
	return 1;
    }
    win_wantscreen = w;
    w->w_cline = 0;
    w->w_ccol = 0;
    w->w_nlines = ti_nlines;
    w->w_ncols = ti_ncols;
    win_open = 1;
    w_refresh();
    wrap.w_func = w_finish;
    wrap.w_next = wraps;
    wraps = &wrap;
    return 0;
}

/*
 * Create a window of given size at given position on screen.
 * Can fail, returns NULL, usual conventions.
 */
static win_t *
new_win(atline, atcol, nlines, ncols)
int	atline;
int	atcol;
int	nlines;
int	ncols;
{
    register win_t	*w;

    if (win_open == 0 && w_init())
	return NULL;

    /*
     * Apply special values of position and size parameters.
     */
    if (atline < 0)
	atline = -atline * (ti_nlines - nlines) / 2;
    if (atcol < 0)
	atcol = -atcol * (ti_ncols - ncols) / 2;
    if (nlines <= 0)
	nlines += ti_nlines;
    if (ncols <= 0)
	ncols += ti_ncols;

    /*
     * Bounce onto the screen, first right and bottom, then left and top.
     */
    if (atline + nlines > ti_nlines)
	atline = ti_nlines - nlines;
    if (atcol + ncols > ti_ncols)
	atcol = ti_ncols - ncols;
    if (atline < 0)
	atline = 0;
    if (atcol < 0)
	atcol = 0;

    /*
     * Crop any remaining overhang.
     */
    if (nlines > ti_nlines)
	nlines = ti_nlines;
    if (ncols > ti_ncols)
	ncols = ti_ncols;

    if ((w = talloc(win_t)) == NULL)
	return NULL;
    objof(w)->o_type = &win_type;
    objof(w)->o_tcode = TC_OTHER;
    objof(w)->o_flags = 0;
    objof(w)->o_nrefs = 1;
    rego(w);
    if ((w->w_image = ici_alloc(nlines * ncols)) == NULL)
	goto fail;
    w->w_nlines = nlines;
    w->w_ncols = ncols;
    w->w_flags |= W_ISDIRTY;
    w->w_mindirt = 0;
    w->w_maxdirt = nlines - 1;
    w->w_atline = atline;
    w->w_atcol = atcol;
    w->w_cline = 0;
    w->w_ccol = 0;
    w->w_up = NULL;
    w->w_down = NULL;
    if (win_bottom == NULL)
    {
	win_cursor = w;
	win_top = w;
	win_bottom = w;
    }
    else
    {
	win_top->w_up = w;
	w->w_down = win_top;
	win_top = w;
    }
    memset(w->w_image, ' ', nlines * ncols);
    return w;

fail:
    if (w != NULL)
	decref(w);
    return NULL;
}

static unsigned long
mark_win(w)
win_t	*w;
{
    objof(w)->o_flags |= O_MARK;
    return sizeof(win_t) + w->w_nlines * w->w_ncols * 2;
}

static void w_pop();

static void
free_win(w)
win_t	*w;
{
    if (w->w_image != NULL)
	w_pop(w);
    ici_free((char *)w);
}

static void
w_pop(w)
register win_t	*w;
{
    if (w == NULL)
	w = win_top;
    if (w->w_image == NULL)
	return;
    if (w == win_bottom)
    {
	if (win_cursor == w)
	    win_cursor = w->w_up;
	if ((win_bottom = w->w_up) == NULL)
	    win_top = NULL;
    }
    if (w == win_top)
    {
	if (win_cursor == w)
	    win_cursor = w->w_down;
	if ((win_top = w->w_down) == NULL)
	    win_bottom = NULL;
    }
    if (w->w_up != NULL)
	w->w_up->w_down = w->w_down;
    if (w->w_down != NULL)
	w->w_down->w_up = w->w_up;
    if (w == win_cursor)
	win_cursor = win_top;
    ici_free(w->w_image);
    w->w_image = NULL;
    win_isamess = 1;
}

static void
w_popall()
{
    while (win_top != NULL)
	w_pop((win_t *)NULL);
}

static void win_goto();

static void
place_cursor()
{
    register int	l;
    register int	c;

    if (win_cursor != NULL)
    {
	if ((l = win_cursor->w_cline + win_cursor->w_atline) < 0)
	    l = 0;
	if (l > ti_nlines - 1)
	    l = ti_nlines - 1;
	if ((c = win_cursor->w_ccol + win_cursor->w_atcol) < 0)
	    c = 0;
	if (c > ti_ncols - 1)
	    c = ti_ncols - 1;
	win_goto(l, c);
    }
    else
	win_goto(ti_nlines - 1, 0);
}

static void w_redraw();

static void
w_refresh()
{
    register win_t	*w;
    char		*startfx;
    char		*endfx;

    if (w_suspended)
    {
#ifndef	MSDOS
	ti_nocanon();
#endif
	w_redraw();
	w_suspended = 0;
    }

    for
    (
	w = win_bottom;
	!win_isamess && w != NULL && !(w->w_flags & W_ISDIRTY);
	w = w->w_up
    )
	;

    if (w == NULL)
    {
	/*
	 * Nothing was dirty.
	 */
	if (win_isamess)
	{
	    screen_clear();
	    memset(win_onscreen->w_image, ' ', ti_nlines * ti_ncols);
	    memset(win_wantscreen->w_image, ' ', ti_nlines * ti_ncols);
	}
	goto fin;
    }

    startfx = NULL;
    endfx = NULL;
    if (win_isamess)
    {
	memset(win_wantscreen->w_image, ' ', ti_nlines * ti_ncols);
	startfx = win_wantscreen->w_image;
	endfx = win_wantscreen->w_image + ti_nlines * ti_ncols;
    }

    while (w != NULL)
    {
	register char	*to;
	register char	*from;
	register int	line;

	/*
	 * Copy the window `w' down onto the screen `win_wantscreen'.
	 * Keep track of the first and last effected bits of the screen
	 * while we are at it.
	 */
	to = w_ptr(win_wantscreen, w->w_atline + w->w_mindirt, w->w_atcol);
	if (startfx == NULL || startfx > to)
	    startfx = to;
	from = w_ptr(w, w->w_mindirt, 0);
	for (line = w->w_mindirt; line <= w->w_maxdirt; ++line)
	{
	    memcpy(to, from, w->w_ncols);
	    to += ti_ncols;
	    from += w->w_ncols;
	}
	to -= ti_ncols;
	to += w->w_ncols;
	if (endfx == NULL || endfx < to)
	    endfx = to;

	w->w_flags &= ~W_ISDIRTY;
	w = w->w_up;
    }

    /*
     * `win_wantscreen' is now how we would like the screen to look.
     * `win_onscreen' is how it does look.  So now we do the "optimal"
     * update.
     */
    {
	register char	*gotc;
	register char	*want;
	register int	i;
	register char	*gotend;
	register int	ncols;

	ncols = ti_ncols;
	want = startfx;
	gotc = win_onscreen->w_image + (want - win_wantscreen->w_image);
	gotend = win_onscreen->w_image + (endfx - win_wantscreen->w_image);
	while (gotc < gotend)
	{
	    if (*gotc++ != *want++)
	    {
		i = gotc - win_onscreen->w_image - 1;
		win_goto(i / ncols, i % ncols);
		win_putc(want[-1]);
		gotc[-1] = want[-1];
	    }
	}
    }
fin:
    place_cursor();
    if (inacs)
    {
    	end_acs();
	inacs = 0;
    }
    fflush(stdout);
    win_isamess = 0;
}

static void
win_goto(line, col)
register int	line;
register int	col;
{
    register char	*p;

    if (!win_lost && line == win_atline)
    {
	if (col == win_atcol)
	     return;
	if (col - win_atcol > 0 && col - win_atcol <= ti_gotocost)
	{
	    p = w_ptr(win_wantscreen, line, win_atcol);
	    while (win_atcol < col)
		win_putc(*p++);
	    return;
	}
    }
    screen_goto(line, col);
    win_atline = line;
    win_atcol = col;
    win_lost = 0;
}

static void
win_putc(c)
int	c;
{
    if (win_atcol == ti_ncols - 1 && win_atline == ti_nlines - 1)
	return;
    c &= 0xFF;
    if (c < ' ' && ti_acsmap[c] != '\0')
    {
	if (!inacs)
	{
	    start_acs();
	    inacs = 1;
	}
	ti_putc(ti_acsmap[c] & 0xFF);
    }
    else
    {
	if (c < ' ')
	    c = ti_defmap[c];
	if (inacs)
	{
	    end_acs();
	    inacs = 0;
	}
	ti_putc(c);
    }

#ifndef	MSDOS
    if (win_atcol == ti_ncols - 1)
    {
	if (ti_automargins)
	{
	    win_atline++;
	    win_atcol = 0;
	}
	else
	    win_lost = 1;
    }
    else
	win_atcol++;
#endif
}

static void
w_cursorat(w, line, col)
register win_t	*w;
int		line;
int		col;
{
    win_cursor = w;
    w->w_cline = line;
    w->w_ccol = col;
}

static void
w_box(w)
register win_t	*w;
{
    register int	i;
    register char	*p;
    register char	*q;

    p = w_ptr(w, 0, 0);
    q = w_ptr(w, w->w_nlines - 1, 0);
    *p++ = CH_ULCORNER;
    *q++ = CH_LLCORNER;
    for (i = 1; i < w->w_ncols - 1; i++)
    {
	*p++ = CH_HLINE;
	*q++ = CH_HLINE;
    }
    *p++ = CH_URCORNER;
    *q++ = CH_LRCORNER;

    p = w_ptr(w, 1, 0);
    q = w_ptr(w, 1, w->w_ncols - 1);
    for (i = 1; i < w->w_nlines - 1; i++)
    {
	*p = CH_VLINE;
	*q = CH_VLINE;
	p += w->w_ncols;
	q += w->w_ncols;
    }
    w_isdirty(w, 0);
    w_isdirty(w, w->w_nlines - 1);
}

static int	paint_text();

win_t *
w_textwin(atline, atcol, text, tabs)
int	atline;
int	atcol;
char	*text;
char	*tabs;
{
    int			ncols;
    int			nlines;
    int			left;
    win_t		*w;

    ncols = 0;
    nlines = 0;
    ncols -= left = paint_text(NULL, &nlines, &ncols, text, tabs);
    if ((w = new_win(atline, atcol, nlines + 2, ncols + 2)) == NULL)
	return NULL;
    w_box(w);
    atline = 1;
    atcol = 1 - left;
    paint_text(w, &atline, &atcol, text, tabs);
    return w;
}

static int
fieldz(p)
register char	*p;
{
    register char	*q;

    for (q = p; *q != '\0' && *q != '\n' && *q != '\t'; ++q)
	;
    return q - p;
}

static void
clippoke(w, line, col, c)
register win_t	*w;
register int	line;
register int	col;
int		c;
{
    if
    (
	w != NULL
	&&
	col >= 0
	&&
	col < w->w_ncols
	&&
	line >= 0
	&&
	line < w->w_nlines
    )
    {
	*w_ptr(w, line, col) = c;
	w_isdirty(w, line);
    }
}

/*
 * Place multiple lines of text on the window `w' at the line and
 * column given with auto-indent on subsequent lines.
 *
 * tabs is an optional (else NULL) tab specification.  It is a string
 * of tab specs, no seperators.  Each tab spec is an optional +, followed
 * by a decimal number followed by an optional leader character, followed
 * by one of L, C or R.  Same semantics as troff.  Last one repeats.
 * If missing, uses standard multiple of 8 columns stuff.  All relative
 * to starting colunm.
 */
static int
paint_text(w, atline, atcol, text, tabs)
register win_t	*w;
int		*atline;
int		*atcol;
register char	*text;
char		*tabs;
{
    register int	col;	/* Current column. */
    register int	maxcol;
    register int	mincol;
    char		*t;	/* Where we are up to in tabs. */
    int			prevts;	/* Previous tab stop. */
    int			leader;

    t = tabs;
    prevts = 0;
    col = *atcol;
    maxcol = col;
    mincol = col;
    while (*text != '\0')
    {
	if (*text == '\t')
	{
	    int		ts;
	    char	*s;
	    int		just = 0;
	    int		ncol;

	    if (t == NULL)
	    {
		while ((++col - *atcol) % 8 != 0)
		    ;
		++text;
		continue;
	    }
	    ++text;
	    if (*t != '+')
		prevts = 0;
	    ts = prevts += ici_strtol(t, &s, 10);
	    leader = '\0';
	    while (*s != '\0')
	    {
		switch (*s)
		{
		default: leader = *s++; continue;
		case 'L': just = 0; break;
		case 'C': just = 1; break;
		case 'R': just = 2; break;
		}
		break;
	    }
	    ncol = *atcol + ts - just * fieldz(text) / 2;
	    if (leader != '\0' && col < ncol)
	    {
		while (col < ncol)
		    clippoke(w, *atline, col++, leader);
	    }
	    col = ncol;
	    if (*s != '\0')
		++s;
	    if (*s != '\0')
		t = s;
	    continue;
	}
	if (*text == '\n')
	{
	    if (text[1] == '\0')
		break;
	    ++*atline;
	    col = *atcol;
	    t = tabs;
	    prevts = 0;
	    ++text;
	    continue;
	}
	clippoke(w, *atline, col, *text++);
	if (col > maxcol)
	    maxcol = col;
	if (col < mincol)
	    mincol = col;
	++col;
    }
    *atcol = maxcol + 1;
    ++*atline;
    return mincol;
}

#ifdef	NOTUSED
static win_t *
w_mcwin(atline, atcol, box, items)
int	atline;
int	atcol;
int	box;
char	**items;
{
    register int	i;
    register int	j;
    int			n;
    int			maxwid;
    int			nlines;
    int			ncols;
    int			columns;
    win_t		*w;

    maxwid = 0;
    for (i = 0; items[i] != NULL; ++i)
    {
	if (maxwid < (j = strlen(items[i])))
	    maxwid = j;
    }
    if (atcol == W_CENTRE)
	ncols = ti_ncols;
    else
	ncols = ti_ncols - atcol;
    columns = (ncols - 2 * box + 1) / (maxwid + 1);
    if (columns > (j = nptrs(items)))
	columns = j;
    ncols = columns * (maxwid + 1) - 1 + 2 * box;
    nlines = (j + columns - 1) / columns + 2 * box;
    if (atline == W_CENTRE)
    {
	if (nlines > ti_nlines)
	    nlines = ti_nlines;
    }
    else
    {
	if (nlines > ti_nlines - atline)
	    nlines = ti_nlines - atline;
    }
    if ((w = new_win(atline, atcol, nlines, ncols)) == NULL)
	return NULL;
    n = 0;
    if (box)
	w_box(w);
    for (i = 0; i < nlines; ++i)
    {
	for (j = 0; j < columns; ++j)
	{
	    if (items[n] == NULL)
		break;
	    strncpy(w_ptr(w, i + box, j * (maxwid+1) + box), items[n], maxwid);
	    ++n;
	}
    }
    return w;
}
#endif

static void w_ungetchar();
static void beep();

static int
edit_field(w, line, col, width)
register win_t	*w;
int		line;
int		col;
int		width;
{
    register int	c;
    register int	n;
    char		*p;

    n = 0;
    for (;;)
    {
	p = w_ptr(w, line, col);
	c = 0;
	while (c < width && buf[c] != '\0')
	    *p++ = buf[c++];
	while (c++ < width)
	    *p++ = ' ';
	w_isdirty(w, line);
	w_cursorat(w, line, col + n);
	if ((p = w_getchar()) == NULL)
	    return 1;
	/*
	if (c == tc_help)
	    help(f->f_help);
	else
	*/
	else if (p[0] == '\b' && p[1] == '\0')
	{
	    if (n > 0)
		buf[--n] = '\0';
	}
	else if (p[1] != '\0' || p[0] < ' ' || p[0] >= 0x7F)
	{
	    w_ungetchar(p);
	    return 0;
	}
	else if (n == width)
	    beep();
	else
	{
	    if (chkbuf(n))
		return 1;
	    buf[n++] = p[0];
	    buf[n] = '\0';
	}
    }
}

static void
w_redraw()
{
    register win_t	*w;

    w = win_onscreen;
    memset(w->w_image, ' ', w->w_nlines * w->w_ncols);
    screen_clear();
    if (win_bottom != NULL)
	w_isdirty(win_bottom, 0);
}

static void
w_ungetchar(s)
char	*s;
{
    strncpy(w_ungotchar, s, sizeof(w_ungotchar));
    w_ungot = 1;
}

static int mesg();

static
char *
w_getchar()
{
    register char	*s;

    if (w_ungot)
    {
	w_ungot = 0;
	return w_ungotchar;
    }

    for (;;)
    {
	if (mesg_transient && mesg_window != NULL)
	    mesg(NULL);
	w_refresh();
#ifndef	MSDOS
	s = ti_getchar();
#else
	w_ungotchar[0] = getch();
	w_ungotchar[1] = '\0';
	s = w_ungotchar;
#endif
	if (s[1] == '\0')
	{
	    if (s[0] == '\r')
		s[0] = '\n';
	    else if (s[0] == '\f')
	    {
		w_redraw();
		continue;
	    }
	}
	if (mesg_window != NULL)
	    mesg(NULL);
	return s;
    }
}

static void
beep()
{
    putchar('\7');
}

#ifdef	NOTUSED
/*
 * Transient message, ie. message that stays up until the start of
 * the next w_getchar().
 */
static void
tmesg(m)
char	*m;
{
    mesg(m);
    mesg_transient = 1;
    w_refresh();
}
#endif

/*
 * Put up a message in a box.
 */
static int
mesg(m)
register char	*m;
{
    register win_t	*w;
    register int	i;

    if (w_init())
	return 1;
    mesg_transient = 0;
    if (mesg_window != NULL)
    {
	w_pop(mesg_window);
	decref(mesg_window);
	mesg_window = NULL;
    }
    if (m != NULL)
    {
	if ((i = strlen(m)) > ti_ncols - 2)
	    i = ti_ncols - 2;
	if ((mesg_window = w = new_win(ti_nlines - 3, W_CENTRE, 3, i + 2)) == NULL)
	    return 1;
	w_box(w);
	strncpy(w_ptr(w, 1, 1), m, i);
	w_cursorat(w, 1, i);
    }
    return 0;
}

static int
assign_win(w, k, v)
register win_t	*w;
object_t	*k;
object_t	*v;
{
    if (isint(k) && isstring(v))
    {
	register int	i;

	if ((i = intof(k)->i_value) < 0 || i >= w->w_nlines * w->w_ncols)
	    return 0;
	w->w_image[i] = stringof(v)->s_chars[0];
	w_isdirty(w, i / w->w_ncols);
	return 0;
    }
    return assign_simple(objof(w), k, v);
}

static object_t *
fetch_win(w, k)
register win_t	*w;
object_t	*k;
{
    if (isint(k))
    {
	register int	i;

	if ((i = intof(k)->i_value) < 0 || i >= w->w_nlines * w->w_ncols)
	    k = objof(new_name("", 0));
	else
	{
	    if ((buf[0] = w->w_image[i]) == '\0')
		buf[0] = ' ';
	    k = objof(new_name(buf, 1));
	}
    }
    else if (k == objof(get_cname("w_nlines")))
	k = objof(new_int((long)w->w_nlines));
    else if (k == objof(get_cname("w_ncols")))
	k = objof(new_int((long)w->w_ncols));
    else if (k == objof(get_cname("w_atline")))
	k = objof(new_int((long)w->w_atline));
    else if (k == objof(get_cname("w_atcol")))
	k = objof(new_int((long)w->w_atline));
    else
	fetch_simple(objof(w), k);
    if (k != NULL)
	decref(k);
    return k;
}

static int
f_w_push()
{
    long	atline;
    long	atcol;
    long	nlines;
    long	ncols;

    if (ici_typecheck("iiii", &atline, &atcol, &nlines, &ncols))
	return 1;
    return ici_ret_with_decref(objof(new_win((int)atline, (int)atcol, (int)nlines, (int)ncols)));
}

static int
f_w_mesg()
{
    char	*s;

    if (ici_typecheck("s", &s))
	return 1;
    if (mesg(s))
	return 1;
    return null_ret();
}

static int
f_w_getchar()
{
    char	*p;

    if (win_open == 0 && w_init())
	return 1;

    if ((p = w_getchar()) == NULL)
	return 1;
    return str_ret(p);
}

static int
f_w_ungetchar()
{
    char	*p;

    if (ici_typecheck("s", &p))
	return 1;
    w_ungetchar(p);
    return null_ret();
}

static int
f_w_box()
{
    win_t	*w;

    if (ici_typecheck("o", &w))
	return 1;
    if (!iswin(objof(w)))
	return ici_argerror(1);
    w_box(w);
    return null_ret();
}

static int
f_w_text()
{
    win_t	*w;
    long	atline;
    long	atcol;
    char	*s;
    char	*tabs;

    if (ici_typecheck("iis*", &atline, &atcol, &s))
	return 1;
    if (NARGS() == 4)
    {
	if (!isstring(ARG(3)))
	    return ici_argerror(3);
	tabs = stringof(ARG(3))->s_chars;
    }
    else
	tabs = NULL;
    if ((w = w_textwin((int)atline, (int)atcol, s, tabs)) == NULL)
	return 1;
    return ici_ret_with_decref(objof(w));
}

static int
f_w_paint()
{
    win_t	*w;
    long	l1;
    long	l2;
    int		atline;
    int		atcol;
    char	*s;
    char	*tabs;

    if (ici_typecheck("oiis*", &w, &l1, &l2, &s))
	return 1;
    atline = l1;
    atcol = l2;
    if (!iswin(objof(w)))
	return ici_argerror(1);
    if (NARGS() == 5)
    {
	if (!isstring(ARG(4)))
	    return ici_argerror(4);
	tabs = stringof(ARG(4))->s_chars;
    }
    else
	tabs = NULL;
    return int_ret((long)paint_text(w, &atline, &atcol, s, tabs));
}

static int
f_w_pop()
{
    win_t	*w;

    if (ici_typecheck("o", &w))
	return 1;
    if (!iswin(objof(w)))
	return ici_argerror(1);
    w_pop(w);
    return null_ret();
}

static int
f_w_clear()
{
    win_t	*w;

    if (ici_typecheck("o", &w))
	return 1;
    if (!iswin(objof(w)))
	return ici_argerror(1);
    memset(w_ptr(w, 0, 0), ' ', w->w_nlines * w->w_ncols);
    w_isdirty(w, 0);
    w_isdirty(w, w->w_nlines - 1);
    return null_ret();
}

static int
f_w_edit()
{
    win_t	*w;
    long	line;
    long	col;
    long	width;
    char	*s;

    if (ici_typecheck("oiiis", &w, &line, &col, &width, &s))
	return 1;
    if (!iswin(objof(w)))
	return ici_argerror(1);
    if (chkbuf(strlen(s)))
	return 1;
    strcpy(buf, s);
    if (edit_field(w, (int)line, (int)col, (int)width))
	return 1;
    return str_ret(buf);
}

static int
f_w_cursorat()
{
    win_t	*w;
    long	line;
    long	col;

    if (ici_typecheck("oii", &w, &line, &col))
	return 1;
    if (!iswin(objof(w)))
	return ici_argerror(1);
    w_cursorat(w, (int)line, (int)col);
    return null_ret();
}

static int
f_w_refresh()
{
    w_refresh();
    return null_ret();
}

static int
f_w_suspend()
{
    w_refresh();
    win_goto(ti_nlines - 1, 0);
    fflush(stdout);
#ifndef	MSDOS
    ti_canon();
#endif
    w_suspended = 1;
    return null_ret();
}

cfunc_t	win_cfuncs[] =
{
    {CF_OBJ,	"w_push",	f_w_push},
    {CF_OBJ,	"w_mesg",	f_w_mesg},
    {CF_OBJ,	"w_getchar",	f_w_getchar},
    {CF_OBJ,	"w_box",	f_w_box},
    {CF_OBJ,	"w_textwin",	f_w_text},
    {CF_OBJ,	"w_pop",	f_w_pop},
    {CF_OBJ,	"w_refresh",	f_w_refresh},
    {CF_OBJ,	"w_paint",	f_w_paint},
    {CF_OBJ,	"w_clear",	f_w_clear},
    {CF_OBJ,	"w_edit",	f_w_edit},
    {CF_OBJ,	"w_cursorat",	f_w_cursorat},
    {CF_OBJ,	"w_suspend",	f_w_suspend},
    {CF_OBJ,	"w_ungetchar",	f_w_ungetchar},
    {CF_OBJ}
};

type_t	win_type =
{
    mark_win,
    free_win,
    hash_unique,
    cmp_unique,
    copy_simple,
    assign_win,
    fetch_win,
    "window"
};
#endif
